-------------------------------------
-- T-SQL Power Tips: Sample Code
-- Richard Broida, Senior Architect, Bennett Adelson
-- rbroida@bennettadelson.com
-------------------------------------

-------------------------------------
-- Tip #3: The OVER Clause
-------------------------------------

-- This query will not execute
--SELECT SalesOrderID,
--		TotalDue,
--		CAST(TotalDue / SUM(TotalDue) * 100. AS NUMERIC(5,2)) AS pct,
--		CAST(TotalDue - AVG(TotalDue) AS NUMERIC(12,2)) AS diff
--FROM [AdventureWorks2008].[Sales].[SalesOrderHeader];

-- This query works, but it makes separate scans of SalesOrderHeader to calculate SUM() and AVG()
SELECT SalesOrderID,
		TotalDue,
		CAST(TotalDue / (SELECT SUM(TotalDue) FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]) * 100. AS NUMERIC(5,2)) AS pct,
		CAST(TotalDue - (SELECT AVG(TotalDue) FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]) AS NUMERIC(12,2)) AS diff
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader];

-- Query with OVER() calcuates SUM() and AVG() in a single scan
SELECT SalesOrderID,
		TotalDue,
		CAST(TotalDue / SUM(TotalDue) OVER() * 100. AS NUMERIC(5,2)) AS pct,
		CAST(TotalDue - AVG(TotalDue) OVER() AS NUMERIC(12,2)) AS diff
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader];

-- Using PARTITION BY to calculate aggregates separately per customer
SELECT SalesOrderID,
		CustomerID,
		TotalDue,
		CAST(TotalDue / SUM(TotalDue) OVER(PARTITION BY CustomerID) * 100. AS NUMERIC(5,2)) AS pct,
		CAST(TotalDue - AVG(TotalDue) OVER(PARTITION BY CustomerID) AS NUMERIC(12,2)) AS diff
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader];

-- Query using cross join instead of OVER(): messy!
SELECT SalesOrderID,
		TotalDue,
		CAST(TotalDue / aggs.GrandTotal * 100. AS NUMERIC(5,2)) AS pct,
		CAST(TotalDue - aggs.Average AS NUMERIC(12,2)) AS diff
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]
CROSS JOIN (SELECT SUM(TotalDue) AS GrandTotal, AVG(TotalDue) AS Average 
	FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]) AS aggs;

-------------------------------------
-- Tip #4: The Four Ranking Functions
-------------------------------------

-- Using the four Ranking functions with OVER
SELECT SalesOrderID, TotalDue,
  ROW_NUMBER() OVER(ORDER BY TotalDue) AS RowNum,
  RANK()       OVER(ORDER BY TotalDue) AS Rnk,
  DENSE_RANK() OVER(ORDER BY TotalDue) AS DenseRnk,
  NTILE(100)     OVER(ORDER BY TotalDue) AS NTile100
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]
ORDER BY TotalDue;

-------------------------------------
-- Tip #5: TOP-less TOPs
-------------------------------------

-- Get the 20 biggest orders, display by OrderDate
WITH Topless AS
(
SELECT SalesOrderID, TotalDue, OrderDate,
  ROW_NUMBER() OVER(ORDER BY TotalDue DESC) AS RowNum
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]
)
SELECT * FROM Topless
WHERE RowNum <= 20
ORDER BY OrderDate;

-- Select every other order per customer
WITH Topless AS
(
SELECT SalesOrderID, CustomerId, TotalDue,
  ROW_NUMBER() OVER(PARTITION BY CustomerId ORDER BY OrderDate, SalesOrderId) AS RowNum
FROM [AdventureWorks2008].[Sales].[SalesOrderHeader]
)
SELECT * FROM Topless
WHERE RowNum % 2 = 0;